#!/usr/bin/env python

#  screenlets.session (c) RYX (aka Rico Pfaus) 2007 <ryx@ryxperience.com>
#
# INFO:
# This module contains the ScreenletSession-class which handles the lower-level
# things like startup, multiple instances and sessions. It should also become
# the interface for load/save operations. The ScreenletSession is further
# responsible for handling command-line args to the Screenlet and should maybe
# offer some convenient way of setting Screenlet-options via commandline (so
# one can do "NotesScreenlet --theme_name=green --scale=0.5" and launch the
# Note with the given theme and scale)..
#
#
# INFO:
# - When a screenlet gets launched:
#   - the first instance of a screenlet creates the Session-object (within the
#     __main__-code)
#   - the session object investigates the config-dir for the given Screenlet
#     and restores available instances
#   - else (if no instance was found) it simply creates a new instance of the 
#     given screenlet and runs its mainloop
# - the --session argument allows setting the name of the session that will be
#   used by the Screenlet (to allow multiple configs for one Screenlet)
#
# TODO:
# - set attributes via commandline??
#

import os
import glob
import random
from xdg import BaseDirectory

import backend			# import screenlets.backend module
import services

# TODO: use XDG instead?
#SESSION_PATH = os.environ['HOME'] + '/.config/Screenlets/'

			
class ScreenletSession (object):
	"""The ScreenletSession manages instances of a Screenlet and handles
	saving/restoring options. Each Screenlet contains a reference to its 
	session. Multiple instances of the same Screenlet share the same
	session-object."""
	
	# constructor
	def __init__ (self, screenlet_classobj, name='default'):
		object.__init__(self)
		# check type
		if not screenlet_classobj.__name__.endswith('Screenlet'):
			# TODO: also check for correct type (Screenlet-subclass)!!
			raise Exception("""ScreenletSession.__init__ has to be called with a
			 valid Screenlet-classobject as first argument!""")
		# init props
		self.name 		= name
		self.screenlet 	= screenlet_classobj
		self.instances 	= []
		# check sys.args for "--session"-argument and override name, if set
		self.__parse_commandline()
		# set session path (and create dir-tree if not existent)
		#self.path 		= SESSION_PATH + screenlet_classobj.__name__[:-9] + \
		#		'/' + self.name + '/'
		p = screenlet_classobj.__name__[:-9] + '/' + self.name + '/'
		self.path = BaseDirectory.load_first_config('Screenlets/' + p)
		if self.path == None:
			self.path = BaseDirectory.save_config_path('Screenlets/' + p)
		if self.path:
			self.backend = backend.CachingBackend(path=self.path)
		else:
			# no config-dir? use dummy-backend and note about problem
			self.backend = backend.ScreenletsBackend()
			print "Unable to init backend - settings will not be saved!"
		"""
		# TEST:
		# init  backend (TODO: create config path if not existent !!!!!!!)
		self.backend	= backend.CachingBackend(path=self.path)
		if self.backend == None:	
			# no backend? use dummy-backend and note about problem
			self.backend = backend.ScreenletsBackend()
			print "Unable to create config ... settings will not be saved!"
		"""
	
	def create_instance (self, id=None, **keyword_args):
		"""Create a new instance with ID 'id' and add it to this session. The 
		function returns either the new Screenlet-instance or None."""
		print "Creating new instance: "
		# if id is none or already exists
		if id==None or id=='' or self.get_instance_by_id(id) != None:
			print "ID is unset or already in use - creating new one!"
			id = self.__get_next_id()
		sl = self.screenlet(id=id, session=self, **keyword_args)
		if sl:
			self.instances.append(sl)		# add screenlet to session
			# and cause initial save to store INI-file in session dir
			sl.x = sl.x
			return sl
		return None
	
	def delete_instance (self, id):
		"""Delete the given instance with ID 'id' and remove its session file.
		When the last insatnce within the session is removed, the session dir 
		is completely removed."""
		sl = self.get_instance_by_id(id)
		if sl:
			# remove instance from session
			self.instances.remove(sl)
			# remove session file
			try:
				self.backend.delete_instance(id)
			except Exception:
				print "Failed to remove INI-file for instance (not critical)."
			# if this was the last instance
			if len(self.instances) == 0:
				# maybe show confirmation popup?
				print "Removing last instance from session"
				# TODO: remove whole session directory
				print "TODO: remove self.path: %s" % self.path
				try:
					os.rmdir(self.path)
				except:
					print "Failed to remove session dir '%s' - not empty?" % self.name
				# ...
				# quit gtk on closing screenlet
				sl.quit_on_close = True
			else:
				print "Removing instance from session but staying alive"
				sl.quit_on_close = False
			# delete screenlet instance
			sl.close()
			del sl
			return True
		return False
	
	def get_instance_by_id (self, id):
		"""Return the instance with the given id from within this session."""
		for inst in self.instances:
			if inst.id == id:
				return inst
		return None
	
	def start (self):
		"""Start a new session (or restore an existing session) for the
		current Screenlet-class. Creates a new instance when none is found.
		Returns True if everything worked well, else False."""
		# check for a running instance first and use dbus-call to add
		# a new instance in that case
		sln = self.screenlet.__name__[:-9]
		if services.service_is_running(sln):
			print "Found a running session - adding new instance by service."""
			srvc = services.get_service_by_name (sln)
			if srvc:
				print srvc
				srvc.add('')
				return False
		# check for existing entries in the session with the given name
		print "Loading instances in: " + self.path
		if self.__load_instances():
			# restored existing entries?
			print "Restored instances from session '%s' ..." % self.name
			# call mainloop of first instance (starts application)
			#self.instances[0].main()
			self.__run_session(self.instances[0])
		else:
			# create new first instance
			print 'No instance(s) found in session-path, creating new one.'
			sl = self.screenlet(session=self, id=self.__get_next_id())
			if sl:
				# add screenlet to session
				self.instances.append(sl)
				# now cause a save of the options to initially create the
				# INI-file for this instance
				self.backend.save_option(sl.id, 'x', sl.x)
				# call mainloop and give control to Screenlet
				#sl.main()
				self.__run_session(sl)
			else:
				print 'Failed creating instance of: ' + self.classobj.__name__
				return False
		# all went well
		return True
	
	def __get_next_id (self):
		"""Get the next ID for an instance of the assigned Screenlet."""
		num = 1
		sln = self.screenlet.__name__[:-9]
		id = sln + str(num)
		while self.get_instance_by_id(id) != None:
			id = sln + str(num)
			num += 1
		return id
		
	def __load_instances (self):
		"""Check for existing instances in the current session, create them 
		and store them into self.instances if any are found. Returns True if 
		at least one instance was found, else False."""
		dirlst = glob.glob(self.path + '*')
		tdlen = len(self.path)
		for filename in dirlst:
			filename = filename[tdlen:]		# strip path from filename
			print 'File: ' + filename
			if filename.endswith('.ini'):
				# create new instance
				sl = self.create_instance(id=filename[:-4], enable_saving=False)
				if sl:
					# set options for the screenlet
					print "Set options in " + sl.__name__
					#self.__restore_options_from_file (sl, self.path + filename)
					self.__restore_options_from_backend(sl, self.path+filename)
					sl.enable_saving(True)
				else:
					print "Failed to create instance of '%s'!" % filename[:-4]
		# if instances were found, return True, else False
		if len(self.instances) > 0:
			return True
		return False
	
	"""
	def __restore_options_from_file (self, screenlet, filename):
		"Restore a screenlet's options from a file and apply them."
		# open file 
		f = None
		try:
			f = open(filename, 'r')
			lines = f.readlines()
			# read all options for this instance from file
			for line in lines:
				#print "LOAD: "+line[:-1]
				# split line and cut trailing EOL
				parts = line[:-1].split('=', 1)
				if len(parts) > 1:
					print 'OPTION: %s = %s' % (parts[0], parts[1])
					opt = screenlet.get_option_by_name(parts[0])
					if opt:
						print opt
						setattr(screenlet, opt.name, opt.on_import(parts[1]))
			f.close()
			return True
		except Exception, ex:
			print "Error while loading options: " + str(ex)
			if f: f.close()
		return False
	"""
	
	# replacement for above function
	def __restore_options_from_backend (self, screenlet, filename):
		"""Restore and apply a screenlet's options from the backend."""
		# disable the canvas-updates in the screenlet
		screenlet.disable_updates = True
		# get options for SL from backend
		opts = self.backend.load_instance(screenlet.id)
		if opts:
			for o in opts:
				# get the attribute's Option-object from Screenlet
				opt = screenlet.get_option_by_name(o)
				# NOTE: set attribute in Screenlet by calling the
				# on_import-function for the Option (to import
				# the value as the required type)
				if opt:
					setattr(screenlet, opt.name, opt.on_import(opts[o]))
		# re-enable updates and call redraw/reshape
		screenlet.disable_updates = False
		screenlet.redraw_canvas()
		screenlet.update_shape()
	
	def __run_session (self, main_instance):
		"""Run the session by calling the main handler of the given Screenlet-
		instance. Handles sigkill (?) and keyboard interrupts."""
		# add sigkill-handler
		import signal
		def on_kill():
			print "Screenlet has been killed. TODO: make this an event"
		signal.signal(signal.SIGTERM, on_kill)
		# start
		try:
			# start mainloop of screenlet
			main_instance.main()
		except KeyboardInterrupt:
			# notify when daemon is closed
			print "Screenlet '%s' has been interrupted by keyboard. TODO: make this an event" % self.screenlet.__name__
		except Exception, ex:
			print "Exception in ScreenletSession: " + ex
	
	def __parse_commandline (self):
		"""Check commandline args for "--session" argument and set session
		name if found. Runs only once during __init__.
		TODO: handle more arguments and maybe allow settings options by
		commandline"""
		import sys
		for arg in sys.argv[1:]:
			# name of session?
			if arg.startswith('--session=') and len(arg)>10:
				self.name = arg[10:]



def create_session (classobj):
	"""A very simple utility-function to easily create/start a new session."""
	session = ScreenletSession(classobj)
	session.start()



# TEST
if __name__ == '__main__':

	# only a dummy - should be a Screenlet-subclass
	class DummyScreenlet:
		__name__ = 'DummyScreenlet'
		def __init__ (self, session=None):
			self.session = session
			self.id = 'Dummy1'
			pass
		def main (self):
			print "Should start screenlet here."
	
	# create (init) new session
	session = ScreenletSession(DummyScreenlet)
	# and start it (i.e. load/create instance(s) ...)
	#session.get_instance_by_id ('Dummy1') 
	session.start()
	
	#print "Session path is: " + session.path
	#print "First session file is: " + session.path + '/0.ini'
	

